// Package sys is an internal package that contains helper methods for dealing
// with Linux that are more complicated than basic wrappers. Basic wrappers
// usually belong in internal/linux. If you feel something belongs in
// libcontainer/utils or libcontainer/system, it probably belongs here instead.
package sys
